import os
import shutil
import ast

import luigi

import lib5c.contrib.luigi.tasks as tasks
from lib5c.util.system import splitall, shell_quote, check_outdir


def directory_to_task(directory, table, all_reps, **kwargs):
    """
    Converts a directory to a TreeMixin Task class instance, using a provided
    table.

    Parameters
    ----------
    directory : str
        The directory identifying this task.
    table : Dict[str, Tuple[str, dict[str, Any]]]
        A map from directory parts to (Task class name, param dict) tuples.
    all_reps : List[str]
        A list of all the replicates.
    kwargs : additional keyword arguments
        Will be passed to the new Task instance. The most common kwarg is 'rep'.

    Returns
    -------
    luigi.Task
        The specified Task instance.
    """
    _, current = os.path.split(directory)
    class_name, params = table[current]
    params = dict(params.items())
    params.update(table=table, directory=directory, all_reps=all_reps)
    params.update(**kwargs)
    instance = globals()[class_name](**params)
    return instance


class TreeMixin(object):
    table = luigi.DictParameter()
    directory = luigi.Parameter()
    outfile_pattern = luigi.Parameter()
    rep = luigi.Parameter(default=None)
    all_reps = luigi.ListParameter()

    def preceding_task(self, rep=None):
        return directory_to_task(os.path.split(self.directory)[0],
                                 self.table, self.all_reps, rep=rep)

    def locus_info_task(self):
        # walk directory to see if you hit something that maps to MakeBinned
        # if you do, return DetermineBins with the indicated bin width
        # otherwise, return PrimerFile
        all_steps = splitall(self.directory)[::-1]
        for step in all_steps:
            if self.table[step][0] == 'MakeBinned':
                bin_width = self.table[step][1].get('bin_width')
                if bin_width is not None:
                    return DetermineBins(bin_width=bin_width)
                else:
                    return DetermineBins()
        return PrimerFile()

    def output(self):
        full_outfile = os.path.join(self.directory, self.outfile_pattern)
        if self.rep is None:
            return luigi.LocalTarget(full_outfile)
        else:
            return luigi.LocalTarget(full_outfile % self.rep)


class PerRepSimpleTreeMixin(TreeMixin):
    def requires(self):
        return self.locus_info_task(), self.preceding_task(rep=self.rep)


class JointTask(TreeMixin, luigi.WrapperTask):
    rep = luigi.Parameter()
    outfile_pattern = luigi.Parameter(default=None)

    def get_inner_task_class(self):
        raise NotImplementedError()

    def get_inner_task_params(self):
        """
        Hook to allow subclasses to supply extra parameters to their inner
        tasks.

        Returns
        -------
        dict
            Extra parameters to be supplied to the inner task upon construction.
        """
        return {}

    def get_inner_task_param_dict(self):
        """
        Constructs the complete dict of params for inner task instantiation.

        Provides some important core defaults in the context of the tree
        pipeline, and injects whatever parameters are returned by
        ``get_inner_task_params()``.

        Returns
        -------
        dict
            The complete dict of params.
        """
        inner_task_params = self.get_inner_task_params()
        inner_task_params.update(table=self.table, directory=self.directory,
                                 all_reps=self.all_reps)
        if hasattr(self, 'outfile_pattern'):
            inner_task_params.update(outfile_pattern=self.outfile_pattern)
        if hasattr(self, 'heatmap'):
            inner_task_params.update(heatmap=self.heatmap)
        if hasattr(self, 'heatmap_outdir'):
            inner_task_params.update(heatmap_outdir=self.heatmap_outdir)
        return inner_task_params

    def get_rep_index(self):
        return self.all_reps.index(self.rep)

    def requires(self):
        return self.get_inner_task_class()(**self.get_inner_task_param_dict())

    def output(self):
        return self.get_inner_task_class()(**self.get_inner_task_param_dict())\
            .output()[self.get_rep_index()]


class JointInnerMixin(object):
    def _match_inputs(self):
        preceding_task = self.preceding_task(rep=self.all_reps[0])
        return os.path.join(preceding_task.directory,
                            preceding_task.outfile_pattern.replace('%s', '*'))

    def requires(self):
        return [self.locus_info_task()] + \
            [self.preceding_task(rep=rep) for rep in self.all_reps]


class JointInnerParallelMixin(JointInnerMixin):
    def output(self):
        full_outfile = os.path.join(self.directory, self.outfile_pattern)
        return [luigi.LocalTarget(full_outfile % rep) for rep in self.all_reps]


class RawCounts(TreeMixin, luigi.ExternalTask):
    rep = luigi.Parameter()
    countsfiles = luigi.DictParameter()
    outfile_pattern = luigi.Parameter(default=None)

    def output(self):
        return luigi.LocalTarget(self.countsfiles[self.rep])


@tasks.visualizable()
class MakeRaw(PerRepSimpleTreeMixin, luigi.Task):
    def run(self):
        check_outdir(self.output().path)
        shutil.copy(self.input()[1].path, self.output().path)


class PrimerFile(luigi.ExternalTask):
    primerfile = luigi.Parameter()

    def output(self):
        return luigi.LocalTarget(self.primerfile)


class DetermineBins(tasks.DetermineBinsTask):
    def requires(self):
        return PrimerFile()

    def output(self):
        return luigi.LocalTarget('bedfiles/%ikb_bins.bed' %
                                 (self.bin_width/1000))


class MakeRemoved(PerRepSimpleTreeMixin, tasks.OutliersTask):
    pass


@tasks.visualizable()
class MakeQnorm(JointTask):
    regional = luigi.BoolParameter(default=False)
    averaging = luigi.BoolParameter(default=False)
    condition_on = luigi.Parameter(default=None)
    reference = luigi.Parameter(default=None)

    def get_inner_task_class(self):
        return QnormInnerTask

    def get_inner_task_params(self):
        return {'regional': self.regional,
                'averaging': self.averaging,
                'condition_on': self.condition_on,
                'reference': self.reference}


class QnormInnerTask(JointInnerParallelMixin, TreeMixin, tasks.QnormTask):
    def _construct_cmd_string(self):
        cmd = 'lib5c qnorm -p %s' % self.input()[0].path
        if self.regional:
            cmd += ' -R'
        if self.averaging:
            cmd += ' -A'
        if self.condition_on is not None:
            cmd += ' -c %s' % shell_quote(self.condition_on)
        if self.reference is not None:
            cmd += ' -r %s' % shell_quote(self.reference)
        cmd += ' %s %s' % (os.path.join(self.directory, self.outfile_pattern),
                           ' '.join(i.path for i in self.input()[1:]))
        return cmd


class MakeExpress(PerRepSimpleTreeMixin, tasks.ExpressTask):
    pass


@tasks.visualizable()
class MakeJointExpress(JointTask):
    def get_inner_task_class(self):
        return JointExpressInnerTask


class JointExpressInnerTask(JointInnerParallelMixin, TreeMixin,
                            tasks.ExpressTask):
    def _construct_cmd_string(self):
        cmd = 'lib5c express -J -p %s' % self.input()[0].path
        if self.bias:
            cmd += ' -B'
        cmd += ' %s %s' % (shell_quote(self._match_inputs()),
                           os.path.join(self.directory, self.outfile_pattern))
        return cmd


class MakeKR(PerRepSimpleTreeMixin, tasks.KnightRuizTask):
    pass


class MakeIced(PerRepSimpleTreeMixin, tasks.IcedTask):
    pass


class MakeSpline(PerRepSimpleTreeMixin, tasks.SplineTask):
    pass


class MakeSmoothed(PerRepSimpleTreeMixin, tasks.SmoothTask):
    pass


class MakeBinned(TreeMixin, tasks.BinTask):
    bin_width = luigi.IntParameter(default=4000)

    def requires(self):
        return (DetermineBins(bin_width=self.bin_width), PrimerFile(),
                self.preceding_task(rep=self.rep))


class MakeExpected(PerRepSimpleTreeMixin, tasks.ExpectedTask):
    pass


class MakeVariance(TreeMixin, tasks.VarianceTask):
    def requires(self):
        exp_task = self.preceding_task(rep=self.rep)
        obs_task = exp_task.preceding_task(rep=self.rep)
        return self.locus_info_task(), obs_task, exp_task


class MakeCrossVariance(TreeMixin, tasks.CrossVarianceTask):
    def requires(self):
        cond = next(c for c in self.conditions.split(',') if c in self.rep)
        exp_task = self.preceding_task(rep=self.rep)
        obs_tasks = [exp_task.preceding_task(rep=r) for r in self.all_reps
                     if cond in r]
        return [self.locus_info_task(), exp_task] + obs_tasks


class MakePvalues(TreeMixin, tasks.PvalueTask):
    def requires(self):
        var_task = self.preceding_task(rep=self.rep)
        exp_task = var_task.preceding_task(rep=self.rep)
        obs_task = exp_task.preceding_task(rep=self.rep)
        return self.locus_info_task(), obs_task, exp_task, var_task


class MakeThreshold(JointInnerMixin, TreeMixin, tasks.ThresholdTask):
    def output(self):
        return (
            luigi.LocalTarget(
                os.path.join(self.directory, self.outfile_pattern)),
            luigi.LocalTarget(
                os.path.join(self.directory, self.dataset_outfile)),
            luigi.LocalTarget(
                os.path.join(self.directory, self.kappa_confusion_outfile))
        )


class MakeQvalues(PerRepSimpleTreeMixin, tasks.QvaluesTask):
    pass


class MakeObsMinusExp(TreeMixin, tasks.SubtractTask):
    def requires(self):
        exp_task = self.preceding_task(rep=self.rep)
        obs_task = exp_task.preceding_task(rep=self.rep)
        return self.locus_info_task(), obs_task, exp_task


class MakeObsOverExp(TreeMixin, tasks.DivideTask):
    def requires(self):
        exp_task = self.preceding_task(rep=self.rep)
        obs_task = exp_task.preceding_task(rep=self.rep)
        return self.locus_info_task(), obs_task, exp_task


class MakeLogged(PerRepSimpleTreeMixin, tasks.LogTask):
    pass


class MakeInteractionScores(PerRepSimpleTreeMixin, tasks.InteractionScoreTask):
    pass


class MakeLegacyPvaluesOne(TreeMixin, tasks.LegacyPvaluesOneTask):
    def requires(self):
        exp_task = self.preceding_task(rep=self.rep)
        obs_task = exp_task.preceding_task(rep=self.rep)
        return self.locus_info_task(), obs_task, exp_task


class PipelineTask(luigi.WrapperTask):
    table = luigi.DictParameter()
    tasks = luigi.ListParameter()

    def requires(self):
        # deduce all_reps
        all_reps = ast.literal_eval(
            luigi.configuration.LuigiConfigParser.instance().get(
                'RawCounts', 'countsfiles')).keys()

        for task in self.tasks:
            # check to see what kind of task this is
            class_name, _ = self.table[os.path.split(task)[1]]
            if class_name == 'MakeThreshold':  # don't parallelize across reps
                yield directory_to_task(task, self.table, all_reps)
            else:
                for rep in all_reps:
                    yield directory_to_task(
                        task, self.table, all_reps, rep=rep)
